
# Programs must be written for people to read, and only incidentally for machines to execute.

# PPT: principal polyhedron triangle...
#
#    0
#   / \
#  1---2

# V: the frequency of the subdivision - don't ask...

# Face vertex numbering for triangulated regular polyhedra - icosa, octa, tetra
#
#                             0                    0,0
#           0                / \                   / \
#          / \              1---4                0,1-1,0
#    V2   1---3       V3   / \ / \    etc        / \ / \
#        / \ / \          2---5---7            0,2-1,1-2,0
#       2---4---5        / \ / \ / \           / \ / \ / \
#                       3---6---8---9        0,3-1,2-2,1-3,0
#
#  which is the flattened version of the non-rectanguar (i, j) array

# Class I

# Method 1 subdivision is the equi-distant division of PPT edges and grid 
# on the face plane with projection to the spherical surface. Chords above 
# PPT edges are not equal. Tends to have a high individual length count.

# Method 2 subdivision is the equi-angular division of PPT edges and arcs 
# between PPT edges. Chords above PPT edges are equal. Generally has a 
# slightly lower length count than method 1.

# Kruschke subdivision allows rows of edges to follow symmetric lesser 
# circles, providing one or more groups of truncation planes. The geometry 
# of most frequencies defines the location of the lesser circle planes.
# V5 and V7 require a mathematical solution to find the location of 
# one plane. The others follow. Has a high length count.

# Mexican subdivision is an enhancement to method 2 where edges in rows parallel to
# PPT edges are equal. Only applies to V4 and V5. Although higher V have only small
# inequalitis in one symmetic group of egdes or one symmetric group of radii.
# Has the smallest possible length count where n = V.

# Class II

# Subdivision of the Schwartz triangle - a right triangle between face centre,
# mid edge and PPT vertex. 6 make a PPT face with reflection / rotations.


include Math

module BHL

  module Polyhedra

    class Icosahedron

      def initialize(radius)

        @radius = radius

        calc_icosa_vertices()

      end # initialize(radius)

      def calc_icosa_vertices()

        vtx_x_1 = @radius * sin(atan(2)) * sin(0.4 * PI)
        vtx_x_2 = @radius * sin(atan(2) / 2)
        vtx_y_1 = @radius * sin(atan(2))
        vtx_y_2 = @radius * sin(atan(2)) * cos(0.4 * PI)
        vtx_y_3 = @radius * sin(atan(2)) * sin(0.3 * PI)
        vtx_z_1 = @radius * sin(atan(2)) / (2 * sin(0.3 * PI))

        @icosa_ppt_vtxs = Array.new(12)

        @icosa_ppt_vtxs[ 0] = [       0.0,       0.0,           @radius]

        @icosa_ppt_vtxs[ 1] = [       0.0, - vtx_y_1, @radius - vtx_z_1]
        @icosa_ppt_vtxs[ 2] = [   vtx_x_1, - vtx_y_2, @radius - vtx_z_1]
        @icosa_ppt_vtxs[ 3] = [   vtx_x_2,   vtx_y_3, @radius - vtx_z_1]
        @icosa_ppt_vtxs[ 4] = [ - vtx_x_2,   vtx_y_3, @radius - vtx_z_1]
        @icosa_ppt_vtxs[ 5] = [ - vtx_x_1, - vtx_y_2, @radius - vtx_z_1]

        @icosa_ppt_vtxs[ 6] = [   vtx_x_2, - vtx_y_3, vtx_z_1 - @radius]
        @icosa_ppt_vtxs[ 7] = [   vtx_x_1,   vtx_y_2, vtx_z_1 - @radius]
        @icosa_ppt_vtxs[ 8] = [       0.0,   vtx_y_1, vtx_z_1 - @radius]
        @icosa_ppt_vtxs[ 9] = [ - vtx_x_1,   vtx_y_2, vtx_z_1 - @radius]
        @icosa_ppt_vtxs[10] = [ - vtx_x_2, - vtx_y_3, vtx_z_1 - @radius]

        @icosa_ppt_vtxs[11] = [       0.0,       0.0,         - @radius]

        # It is good to have a component instance axis plane parallel to the icosa
        # component face if one wants to flesh out the edges with substance later. 
        # And also, if the component is basically in one plane it is easier 
        # to select and looks better if the bounding box follows that plane.

        # Three options as best I can see / have found.

        # All use icosa mid front face PPT vertices 1, 6, 10 as the reference face.

        # 1/. Draw on mid front face and only apply icosa face transforms
        # 2/. Draw on mid front face and apply single transform to XZ plane, then 
        #     icosa transforms and finally a component definition entities transform.
        # 3/. Draw on the plane parallel to XZ offset on - Y to face centre and 
        #     apply a single transform back to icosa mid front face plane and
        #     then the icosa face transforms.
        # 4/. Is there another way???

        # 1/. Bounding box is ok with mid front face but axes not aligned with icosa face
        #     Other polyhedra don't have a face aligned so close to XZ plane
        # 2/. Some initial bounding box weirdness. Open the component for editing, 
        #     escape out and bounding box is as should be. New component instances 
        #     are not aligned with model. Axes are Ok
        # 3/. No bounding box weirdness. New component instances are not aligned with 
        #     model. Axes OK - this is the one to use especially for other polyhedra. 

        # Things would be a lot better if we could apply a transform to a component 
        # definition. Or at least mimick SU behaviour when setting axes in components 
        # and applying to all instances...


        # This transform brings the icosa ppt 1, 6, 10 (mid front face) 
        # vertices to the plane parallel to XZ 

        @trnsfrm_angle =

          PI / 2 - (2 * asin(sin(atan(2)) * sin(2 * PI / 5)) - 
                        atan(tan(atan(2) / 2) / cos(PI / 5)))

        trnsfrm_face_vtxs = Geom::Transformation.new(ORIGIN, X_AXIS, - @trnsfrm_angle)

        @icosa_face_vtxs = Array.new(5)

#        @cmpnt_draw_optns = 1
#        @cmpnt_draw_optns = 2
        @cmpnt_draw_optns = 3

        # anti clockwise gives front face out

        @icosa_face_vtxs[0] = @icosa_ppt_vtxs[ 1]
        @icosa_face_vtxs[1] = @icosa_ppt_vtxs[10]
        @icosa_face_vtxs[2] = @icosa_ppt_vtxs[ 6]

        # face centre

        @icosa_face_vtxs[3] = BHL::Polyhedra.get_3_vtx_avg_prjct(@icosa_ppt_vtxs[ 1], 
                                                                 @icosa_ppt_vtxs[10], 
                                                                 @icosa_ppt_vtxs[ 6], @radius)

        # use this point for local zaxis, valid under transform

        @icosa_face_vtxs[4] = @icosa_ppt_vtxs[0]

        if (@cmpnt_draw_optns == 3)

          @icosa_face_vtxs[0] = @icosa_ppt_vtxs[ 1].transform(trnsfrm_face_vtxs)
          @icosa_face_vtxs[1] = @icosa_ppt_vtxs[10].transform(trnsfrm_face_vtxs)
          @icosa_face_vtxs[2] = @icosa_ppt_vtxs[ 6].transform(trnsfrm_face_vtxs)

          @icosa_face_vtxs[3].transform!(trnsfrm_face_vtxs)
          @icosa_face_vtxs[4].transform!(trnsfrm_face_vtxs)

        end

      end # calc_icosa_vertices(radius)

      def get_icosa_ppt_vertices()

        @icosa_ppt_vtxs

      end # get_icosa_ppt_vertices()

      def get_icosa_face_vertices()

        @icosa_face_vtxs

      end # get_icosa_face_vertices()

      def create_face_component(model)

        cmpnt_def = model.definitions.add

        cmpnt_def.name = "Icosahedron V1 Face"

        cmpnt_def.entities.add_face(@icosa_face_vtxs[0], @icosa_face_vtxs[1], @icosa_face_vtxs[2])

        cmpnt_def

      end # create_face_component(model)

      def add_face_components(model, cmpnt_def, all_faces)

        # A little trickery. There are 20 faces so there could be 20 transforms.
        # But a specific set of 4 faces will have the same 5 transforms - the initial instances and 4 more...
        # Transforms[0..3] give one set of 4 such faces. Transforms[4..7] apply to subsequent instances

        trnsfrms_ppt_vtxs = [

          Geom::Transformation.new(ORIGIN),
          Geom::Transformation.new(ORIGIN, @icosa_ppt_vtxs[ 6], 2 * PI / 5),
          Geom::Transformation.new(ORIGIN, @icosa_ppt_vtxs[ 6], 4 * PI / 5),
          Geom::Transformation.new(ORIGIN, @icosa_ppt_vtxs[10], 2 * PI / 5),
          Geom::Transformation.new(ORIGIN, @icosa_ppt_vtxs[ 7], 2 * PI / 5),
          Geom::Transformation.new(ORIGIN, @icosa_ppt_vtxs[ 7], 4 * PI / 5),
          Geom::Transformation.new(ORIGIN, @icosa_ppt_vtxs[ 7], 6 * PI / 5),
          Geom::Transformation.new(ORIGIN, @icosa_ppt_vtxs[ 7], 8 * PI / 5)]

        # This transform aligns the icosa mid front face (defined by 
        # @icosa_ppt_vtxs 1, 6, 10) parallel to the XZ plane 

        trnsfrm_face = Geom::Transformation.new(ORIGIN, X_AXIS, @trnsfrm_angle)

        group = model.entities.add_group()

#        all_faces = false

        if all_faces

          for idx in 0..3

            if (@cmpnt_draw_optns == 1)

              group_inst = group.entities.add_instance(cmpnt_def, trnsfrms_ppt_vtxs[0])

            elsif ((@cmpnt_draw_optns == 2) || (@cmpnt_draw_optns == 3))

              group_inst = group.entities.add_instance(cmpnt_def, trnsfrm_face)

            end

            group_inst.transform!(trnsfrms_ppt_vtxs[idx])

            for jdx in 0..3

              if (@cmpnt_draw_optns == 1)

                group_inst = group.entities.add_instance(cmpnt_def, trnsfrms_ppt_vtxs[0])

              elsif ((@cmpnt_draw_optns == 2) || (@cmpnt_draw_optns == 3))

                group_inst = group.entities.add_instance(cmpnt_def, trnsfrm_face)

              end

              group_inst.transform!(trnsfrms_ppt_vtxs[idx])

              group.entities.transform_entities(trnsfrms_ppt_vtxs[jdx + 4], group_inst)

            end

          end

        else

          # only one face instance

          if (@cmpnt_draw_optns == 1)

            group_inst = group.entities.add_instance(cmpnt_def, trnsfrms_ppt_vtxs[0])

          elsif ((@cmpnt_draw_optns == 2) || (@cmpnt_draw_optns == 3))

            group_inst = group.entities.add_instance(cmpnt_def, trnsfrm_face)

          end

        end

        if (@cmpnt_draw_optns == 2)

          cmpnt_def.entities.transform_entities(trnsfrm_face.inverse, cmpnt_def.entities.to_a)

        end

        group

      end # add_face_components(model, cmpnt_def, all_faces)

      @@face_vtx_map_v3 = [[0, 1, 4], [1, 2, 5], [2, 3, 6],
                           [1, 4, 5], [2, 5, 6], [4, 5, 7],
                           [5, 6, 8], [5, 7, 8], [7, 8, 9]]

      @@face_vtx_map_v4 = [[ 0,  1,  5], [ 1,  2,  6], [ 2,  3,  7], [ 3,  4,  8],
                           [ 1,  5,  6], [ 2,  6,  7], [ 3,  7,  8], [ 5,  6,  9],
                           [ 6,  7, 10], [ 7,  8, 11], [ 6,  9, 10], [ 7, 10, 11],
                           [ 9, 10, 12], [10, 11, 13], [10, 12, 13], [12, 13, 14]]

      @@face_vtx_map_v5 = [[ 0,  1,  6], [ 1,  2,  7], [ 2,  3,  8], [ 3,  4,  9], [ 4,  5, 10],
                           [ 1,  6,  7], [ 2,  7,  8], [ 3,  8,  9], [ 4,  9, 10], [ 6,  7, 11],
                           [ 7,  8, 12], [ 8,  9, 13], [ 9, 10, 14], [ 7, 11, 12], [ 8, 12, 13],
                           [ 9, 13, 14], [11, 12, 15], [12, 13, 16], [13, 14, 17], [12, 15, 16],
                           [13, 16, 17], [15, 16, 18], [16, 17, 19], [16, 18, 19], [18, 19, 20]]

      @@face_vtx_map_v6 = [[ 0,  1,  7], [ 1,  2,  8], [ 2,  3,  9], [ 3,  4, 10], [ 4,  5, 11], [ 5,  6, 12],
                           [ 1,  7,  8], [ 2,  8,  9], [ 3,  9, 10], [ 4, 10, 11], [ 5, 11, 12], [ 7,  8, 13], 
                           [ 8,  9, 14], [ 9, 10, 15], [10, 11, 16], [11, 12, 17], [ 8, 13, 14], [ 9, 14, 15], 
                           [10, 15, 16], [11, 16, 17], [13, 14, 18], [14, 15, 19], [15, 16, 20], [16, 17, 21],
                           [14, 18, 19], [15, 19, 20], [16, 20, 21], [18, 19, 22], [19, 20, 23], [20, 21, 24],
                           [19, 22, 23], [20, 23, 24], [22, 23, 25], [23, 24, 26], [23, 25, 26], [25, 26, 27]]

    end # class Icosahedron

    class IcosahedronCIV2 < Icosahedron

      def initialize(radius)

        super

        @icosa_vtxs = get_icosa_face_vertices

        calc_icosa_civ2_vertices()

      end # initialize(radius)

      def calc_icosa_civ2_vertices()

        @icosa_civ2_vtxs = Array.new(6)

        @icosa_civ2_vtxs[0] = @icosa_vtxs[0]
        @icosa_civ2_vtxs[2] = @icosa_vtxs[1]
        @icosa_civ2_vtxs[5] = @icosa_vtxs[2]

        @icosa_civ2_vtxs[1] = BHL::Polyhedra.get_linear_cmbntn_prjct(@icosa_civ2_vtxs[0], @icosa_civ2_vtxs[2], 1, 2, @radius)
        @icosa_civ2_vtxs[3] = BHL::Polyhedra.get_linear_cmbntn_prjct(@icosa_civ2_vtxs[0], @icosa_civ2_vtxs[5], 1, 2, @radius)
        @icosa_civ2_vtxs[4] = BHL::Polyhedra.get_linear_cmbntn_prjct(@icosa_civ2_vtxs[2], @icosa_civ2_vtxs[5], 1, 2, @radius)

      end # calc_icosa_civ2_vertices()

      def get_icosa_civ2_vertices()

        @icosa_civ2_vtxs

      end # get_icosa_civ2_vertices()

      def create_civ2_face_component(model)

        cmpnt_def = model.definitions.add

        cmpnt_def.name = "Icosahedron Class I V2 Face"

        cmpnt_def.entities.add_face(@icosa_civ2_vtxs[0], @icosa_civ2_vtxs[1], @icosa_civ2_vtxs[3])
        cmpnt_def.entities.add_face(@icosa_civ2_vtxs[1], @icosa_civ2_vtxs[2], @icosa_civ2_vtxs[4])
        cmpnt_def.entities.add_face(@icosa_civ2_vtxs[1], @icosa_civ2_vtxs[3], @icosa_civ2_vtxs[4])
        cmpnt_def.entities.add_face(@icosa_civ2_vtxs[3], @icosa_civ2_vtxs[4], @icosa_civ2_vtxs[5])

        cmpnt_def

      end # create_civ2_face_component(model)

    end # class IcosahedronCIV2

    class IcosahedronCIM1V3 < Icosahedron

      def initialize(radius)

        super

        @icosa_vtxs = get_icosa_face_vertices

        calc_icosa_cim1v3_vertices()

      end # initialize(radius)

      def calc_icosa_cim1v3_vertices()

        @icosa_cim1v3_vtxs = Array.new(10)

        @icosa_cim1v3_vtxs[0] = @icosa_vtxs[0]
        @icosa_cim1v3_vtxs[3] = @icosa_vtxs[1]
        @icosa_cim1v3_vtxs[9] = @icosa_vtxs[2]

        @icosa_cim1v3_vtxs[5] = @icosa_vtxs[3]

        @icosa_cim1v3_vtxs[1] = BHL::Polyhedra.get_linear_cmbntn_prjct(@icosa_cim1v3_vtxs[0], @icosa_cim1v3_vtxs[3], 1, 3, @radius)
        @icosa_cim1v3_vtxs[2] = BHL::Polyhedra.get_linear_cmbntn_prjct(@icosa_cim1v3_vtxs[3], @icosa_cim1v3_vtxs[0], 1, 3, @radius)
        @icosa_cim1v3_vtxs[4] = BHL::Polyhedra.get_linear_cmbntn_prjct(@icosa_cim1v3_vtxs[0], @icosa_cim1v3_vtxs[9], 1, 3, @radius)
        @icosa_cim1v3_vtxs[6] = BHL::Polyhedra.get_linear_cmbntn_prjct(@icosa_cim1v3_vtxs[3], @icosa_cim1v3_vtxs[9], 1, 3, @radius)
        @icosa_cim1v3_vtxs[7] = BHL::Polyhedra.get_linear_cmbntn_prjct(@icosa_cim1v3_vtxs[9], @icosa_cim1v3_vtxs[0], 1, 3, @radius)
        @icosa_cim1v3_vtxs[8] = BHL::Polyhedra.get_linear_cmbntn_prjct(@icosa_cim1v3_vtxs[9], @icosa_cim1v3_vtxs[3], 1, 3, @radius)

      end # calc_icosa_cim1v3_vertices()

      def get_icosa_cim1v3_vertices()

        @icosa_cim1v3_vtxs

      end # get_icosa_cim1v3_vertices()

      def create_cim1v3_face_component(model)

        cmpnt_def = model.definitions.add

        cmpnt_def.name = "Icosa C I Method 1 V3 Face"

        @@face_vtx_map_v3.each do |face|

          cmpnt_def.entities.add_face(@icosa_cim1v3_vtxs[face[0]], @icosa_cim1v3_vtxs[face[1]], @icosa_cim1v3_vtxs[face[2]])

        end

        cmpnt_def

      end # create_cim1v3_face_component(model)

    end # class IcosahedronCIM1V3

    class IcosahedronCIM1V4 < IcosahedronCIV2

      def initialize(radius)

        super

        @icosa_civ2_vtxs = get_icosa_civ2_vertices

        calc_icosa_cim1v4_vertices()

      end # initialize(radius)

      def calc_icosa_cim1v4_vertices()

        @icosa_cim1v4_vtxs = Array.new(15)

        @icosa_cim1v4_vtxs[ 0] = @icosa_civ2_vtxs[0]
        @icosa_cim1v4_vtxs[ 2] = @icosa_civ2_vtxs[1]
        @icosa_cim1v4_vtxs[ 4] = @icosa_civ2_vtxs[2]
        @icosa_cim1v4_vtxs[ 9] = @icosa_civ2_vtxs[3]
        @icosa_cim1v4_vtxs[11] = @icosa_civ2_vtxs[4]
        @icosa_cim1v4_vtxs[14] = @icosa_civ2_vtxs[5]

        @icosa_cim1v4_vtxs[ 1] = BHL::Polyhedra.get_linear_cmbntn_prjct(@icosa_cim1v4_vtxs[0], @icosa_cim1v4_vtxs[4], 1, 4, @radius)
        @icosa_cim1v4_vtxs[ 3] = BHL::Polyhedra.get_linear_cmbntn_prjct(@icosa_cim1v4_vtxs[4], @icosa_cim1v4_vtxs[0], 1, 4, @radius)
        @icosa_cim1v4_vtxs[ 5] = BHL::Polyhedra.get_linear_cmbntn_prjct(@icosa_cim1v4_vtxs[0], @icosa_cim1v4_vtxs[14], 1, 4, @radius)
        @icosa_cim1v4_vtxs[ 8] = BHL::Polyhedra.get_linear_cmbntn_prjct(@icosa_cim1v4_vtxs[4], @icosa_cim1v4_vtxs[14], 1, 4, @radius)
        @icosa_cim1v4_vtxs[12] = BHL::Polyhedra.get_linear_cmbntn_prjct(@icosa_cim1v4_vtxs[14], @icosa_cim1v4_vtxs[0], 1, 4, @radius)
        @icosa_cim1v4_vtxs[13] = BHL::Polyhedra.get_linear_cmbntn_prjct(@icosa_cim1v4_vtxs[14], @icosa_cim1v4_vtxs[4], 1, 4, @radius)

        @icosa_cim1v4_vtxs[ 6] = BHL::Polyhedra.get_linear_cmbntn_prjct(@icosa_cim1v4_vtxs[2], @icosa_cim1v4_vtxs[9], 1, 2, @radius)
        @icosa_cim1v4_vtxs[ 7] = BHL::Polyhedra.get_linear_cmbntn_prjct(@icosa_cim1v4_vtxs[2], @icosa_cim1v4_vtxs[11], 1, 2, @radius)
        @icosa_cim1v4_vtxs[10] = BHL::Polyhedra.get_linear_cmbntn_prjct(@icosa_cim1v4_vtxs[9], @icosa_cim1v4_vtxs[11], 1, 2, @radius)

      end # calc_icosa_cim1v4_vertices()

      def get_icosa_cim1v4_vertices()

        @icosa_cim1v4_vtxs

      end # get_icosa_cim1v4_vertices()

      def create_cim1v4_face_component(model)

        cmpnt_def = model.definitions.add

        cmpnt_def.name = "Icosa C I Method 1 V4 Face"

        @@face_vtx_map_v4.each do |face|

          cmpnt_def.entities.add_face(@icosa_cim1v4_vtxs[face[0]], @icosa_cim1v4_vtxs[face[1]], @icosa_cim1v4_vtxs[face[2]])

        end

        cmpnt_def

      end # create_cim1v4_face_component(model)

    end # class IcosahedronCIV4

    class IcosahedronCIM1V5 < Icosahedron

      def initialize(radius)

        super

        @icosa_vtxs = get_icosa_face_vertices

        calc_icosa_cim1v5_vertices()

      end # initialize(radius)

      def calc_icosa_cim1v5_vertices()

        @icosa_cim1v5_vtxs = Array.new(21)

        @icosa_cim1v5_vtxs[ 0] = @icosa_vtxs[0]
        @icosa_cim1v5_vtxs[ 5] = @icosa_vtxs[1]
        @icosa_cim1v5_vtxs[20] = @icosa_vtxs[2]

        @icosa_cim1v5_vtxs[ 1] = BHL::Polyhedra.get_linear_cmbntn_prjct(@icosa_cim1v5_vtxs[0], @icosa_cim1v5_vtxs[5], 1, 5, @radius)
        @icosa_cim1v5_vtxs[ 4] = BHL::Polyhedra.get_linear_cmbntn_prjct(@icosa_cim1v5_vtxs[5], @icosa_cim1v5_vtxs[0], 1, 5, @radius)
        @icosa_cim1v5_vtxs[ 6] = BHL::Polyhedra.get_linear_cmbntn_prjct(@icosa_cim1v5_vtxs[0], @icosa_cim1v5_vtxs[20], 1, 5, @radius)
        @icosa_cim1v5_vtxs[10] = BHL::Polyhedra.get_linear_cmbntn_prjct(@icosa_cim1v5_vtxs[5], @icosa_cim1v5_vtxs[20], 1, 5, @radius)
        @icosa_cim1v5_vtxs[18] = BHL::Polyhedra.get_linear_cmbntn_prjct(@icosa_cim1v5_vtxs[20], @icosa_cim1v5_vtxs[0], 1, 5, @radius)
        @icosa_cim1v5_vtxs[19] = BHL::Polyhedra.get_linear_cmbntn_prjct(@icosa_cim1v5_vtxs[20], @icosa_cim1v5_vtxs[5], 1, 5, @radius)

        @icosa_cim1v5_vtxs[ 2] = BHL::Polyhedra.get_linear_cmbntn_prjct(@icosa_cim1v5_vtxs[0], @icosa_cim1v5_vtxs[5], 2, 5, @radius)
        @icosa_cim1v5_vtxs[ 3] = BHL::Polyhedra.get_linear_cmbntn_prjct(@icosa_cim1v5_vtxs[5], @icosa_cim1v5_vtxs[0], 2, 5, @radius)
        @icosa_cim1v5_vtxs[11] = BHL::Polyhedra.get_linear_cmbntn_prjct(@icosa_cim1v5_vtxs[0], @icosa_cim1v5_vtxs[20], 2, 5, @radius)
        @icosa_cim1v5_vtxs[14] = BHL::Polyhedra.get_linear_cmbntn_prjct(@icosa_cim1v5_vtxs[5], @icosa_cim1v5_vtxs[20], 2, 5, @radius)
        @icosa_cim1v5_vtxs[15] = BHL::Polyhedra.get_linear_cmbntn_prjct(@icosa_cim1v5_vtxs[20], @icosa_cim1v5_vtxs[0], 2, 5, @radius)
        @icosa_cim1v5_vtxs[17] = BHL::Polyhedra.get_linear_cmbntn_prjct(@icosa_cim1v5_vtxs[20], @icosa_cim1v5_vtxs[5], 2, 5, @radius)

        @icosa_cim1v5_vtxs[ 7] = BHL::Polyhedra.get_linear_cmbntn_prjct(@icosa_cim1v5_vtxs[2], @icosa_cim1v5_vtxs[11], 1, 2, @radius)
        @icosa_cim1v5_vtxs[ 9] = BHL::Polyhedra.get_linear_cmbntn_prjct(@icosa_cim1v5_vtxs[3], @icosa_cim1v5_vtxs[14], 1, 2, @radius)
        @icosa_cim1v5_vtxs[16] = BHL::Polyhedra.get_linear_cmbntn_prjct(@icosa_cim1v5_vtxs[15], @icosa_cim1v5_vtxs[17], 1, 2, @radius)

        @icosa_cim1v5_vtxs[ 8] = BHL::Polyhedra.get_linear_cmbntn_prjct(@icosa_cim1v5_vtxs[3], @icosa_cim1v5_vtxs[15], 1, 3, @radius)
        @icosa_cim1v5_vtxs[12] = BHL::Polyhedra.get_linear_cmbntn_prjct(@icosa_cim1v5_vtxs[11], @icosa_cim1v5_vtxs[14], 1, 3, @radius)
        @icosa_cim1v5_vtxs[13] = BHL::Polyhedra.get_linear_cmbntn_prjct(@icosa_cim1v5_vtxs[14], @icosa_cim1v5_vtxs[11], 1, 3, @radius)

      end # calc_icosa_cim1v5_vertices()

      def get_icosa_cim1v5_vertices()

        @icosa_cim1v5_vtxs

      end # get_icosa_cim1v5_vertices()

      def create_cim1v5_face_component(model)

        cmpnt_def = model.definitions.add

        cmpnt_def.name = "Icosa C I Method 1 V5 Face"

        @@face_vtx_map_v5.each do |face|

          cmpnt_def.entities.add_face(@icosa_cim1v5_vtxs[face[0]], @icosa_cim1v5_vtxs[face[1]], @icosa_cim1v5_vtxs[face[2]])

        end

        cmpnt_def

      end # create_cim1v5_face_component(model)

    end # class IcosahedronCIV5

    class IcosahedronCIM1V6 < Icosahedron

      def initialize(radius)

        super

        @icosa_civ2 = IcosahedronCIV2.new(radius)

        @icosa_civ2_vtxs = @icosa_civ2.get_icosa_civ2_vertices

        @icosa_cim1v3 = IcosahedronCIM1V3.new(radius)

        @icosa_cim1v3_vtxs = @icosa_cim1v3.get_icosa_cim1v3_vertices

        calc_icosa_cim1v6_vertices()

      end # initialize(radius)

      def calc_icosa_cim1v6_vertices()

        @icosa_cim1v6_vtxs = Array.new(28)

        @icosa_cim1v6_vtxs[ 0] = @icosa_civ2_vtxs[0]
        @icosa_cim1v6_vtxs[ 2] = @icosa_cim1v3_vtxs[1]
        @icosa_cim1v6_vtxs[ 3] = @icosa_civ2_vtxs[1]
        @icosa_cim1v6_vtxs[ 4] = @icosa_cim1v3_vtxs[2]
        @icosa_cim1v6_vtxs[ 6] = @icosa_civ2_vtxs[2]
        @icosa_cim1v6_vtxs[13] = @icosa_cim1v3_vtxs[4]
        @icosa_cim1v6_vtxs[15] = @icosa_cim1v3_vtxs[5]
        @icosa_cim1v6_vtxs[17] = @icosa_cim1v3_vtxs[6]
        @icosa_cim1v6_vtxs[18] = @icosa_civ2_vtxs[3]
        @icosa_cim1v6_vtxs[21] = @icosa_civ2_vtxs[4]
        @icosa_cim1v6_vtxs[22] = @icosa_cim1v3_vtxs[7]
        @icosa_cim1v6_vtxs[24] = @icosa_cim1v3_vtxs[8]
        @icosa_cim1v6_vtxs[27] = @icosa_civ2_vtxs[5]

        @icosa_cim1v6_vtxs[ 1] = BHL::Polyhedra.get_linear_cmbntn_prjct(@icosa_cim1v6_vtxs[0], @icosa_cim1v6_vtxs[6], 1, 6, @radius)
        @icosa_cim1v6_vtxs[ 5] = BHL::Polyhedra.get_linear_cmbntn_prjct(@icosa_cim1v6_vtxs[6], @icosa_cim1v6_vtxs[0], 1, 6, @radius)
        @icosa_cim1v6_vtxs[ 7] = BHL::Polyhedra.get_linear_cmbntn_prjct(@icosa_cim1v6_vtxs[0], @icosa_cim1v6_vtxs[27], 1, 6, @radius)
        @icosa_cim1v6_vtxs[12] = BHL::Polyhedra.get_linear_cmbntn_prjct(@icosa_cim1v6_vtxs[6], @icosa_cim1v6_vtxs[27], 1, 6, @radius)
        @icosa_cim1v6_vtxs[25] = BHL::Polyhedra.get_linear_cmbntn_prjct(@icosa_cim1v6_vtxs[27], @icosa_cim1v6_vtxs[0], 1, 6, @radius)
        @icosa_cim1v6_vtxs[26] = BHL::Polyhedra.get_linear_cmbntn_prjct(@icosa_cim1v6_vtxs[27], @icosa_cim1v6_vtxs[6], 1, 6, @radius)

        @icosa_cim1v6_vtxs[ 8] = BHL::Polyhedra.get_linear_cmbntn_prjct(@icosa_cim1v6_vtxs[2], @icosa_cim1v6_vtxs[13], 1, 2, @radius)
        @icosa_cim1v6_vtxs[11] = BHL::Polyhedra.get_linear_cmbntn_prjct(@icosa_cim1v6_vtxs[4], @icosa_cim1v6_vtxs[17], 1, 2, @radius)
        @icosa_cim1v6_vtxs[23] = BHL::Polyhedra.get_linear_cmbntn_prjct(@icosa_cim1v6_vtxs[22], @icosa_cim1v6_vtxs[24], 1, 2, @radius)

        @icosa_cim1v6_vtxs[ 9] = BHL::Polyhedra.get_linear_cmbntn_prjct(@icosa_cim1v6_vtxs[3], @icosa_cim1v6_vtxs[18], 1, 3, @radius)
        @icosa_cim1v6_vtxs[10] = BHL::Polyhedra.get_linear_cmbntn_prjct(@icosa_cim1v6_vtxs[3], @icosa_cim1v6_vtxs[21], 1, 3, @radius)
        @icosa_cim1v6_vtxs[14] = BHL::Polyhedra.get_linear_cmbntn_prjct(@icosa_cim1v6_vtxs[18], @icosa_cim1v6_vtxs[3], 1, 3, @radius)
        @icosa_cim1v6_vtxs[16] = BHL::Polyhedra.get_linear_cmbntn_prjct(@icosa_cim1v6_vtxs[21], @icosa_cim1v6_vtxs[3], 1, 3, @radius)
        @icosa_cim1v6_vtxs[19] = BHL::Polyhedra.get_linear_cmbntn_prjct(@icosa_cim1v6_vtxs[18], @icosa_cim1v6_vtxs[21], 1, 3, @radius)
        @icosa_cim1v6_vtxs[20] = BHL::Polyhedra.get_linear_cmbntn_prjct(@icosa_cim1v6_vtxs[21], @icosa_cim1v6_vtxs[18], 1, 3, @radius)

      end # calc_icosa_cim1v6_vertices()

      def get_icosa_cim1v6_vertices()

        @icosa_cim1v6_vtxs

      end # get_icosa_cim1v6_vertices()

      def create_cim1v6_face_component(model)

        cmpnt_def = model.definitions.add

        cmpnt_def.name = "Icosa C I Method 1 V6 Face"

        @@face_vtx_map_v6.each do |face|

          cmpnt_def.entities.add_face(@icosa_cim1v6_vtxs[face[0]], @icosa_cim1v6_vtxs[face[1]], @icosa_cim1v6_vtxs[face[2]])

        end

        cmpnt_def

      end # create_cim1v6_face_component(model)

      def add_cim1v6_face_components(model, cmpnt_def, all_faces)

        add_face_components(model, cmpnt_def, all_faces)

      end # add_cim1v6_face_components(model, cmpnt_def, all_faces)

    end # class IcosahedronCIM1V6

    class IcosahedronCIM2V3 < Icosahedron

      def initialize(radius)

        super

        @icosa_vtxs = get_icosa_face_vertices

        calc_icosa_cim2v3_vertices()

      end # initialize(radius)

      def calc_icosa_cim2v3_vertices()

        @icosa_cim2v3_vtxs = Array.new(10)

        @icosa_cim2v3_vtxs[0] = @icosa_vtxs[0]
        @icosa_cim2v3_vtxs[3] = @icosa_vtxs[1]
        @icosa_cim2v3_vtxs[9] = @icosa_vtxs[2]

        @icosa_cim2v3_vtxs[5] = @icosa_vtxs[3]

        mid_vtxs = BHL::Polyhedra.get_arc_mid_vtxs_prjct(@icosa_cim2v3_vtxs[0], @icosa_cim2v3_vtxs[3], 3, @radius)

        @icosa_cim2v3_vtxs[1] = mid_vtxs[0]
        @icosa_cim2v3_vtxs[2] = mid_vtxs[1]

        mid_vtxs = BHL::Polyhedra.get_arc_mid_vtxs_prjct(@icosa_cim2v3_vtxs[0], @icosa_cim2v3_vtxs[9], 3, @radius)

        @icosa_cim2v3_vtxs[4] = mid_vtxs[0]
        @icosa_cim2v3_vtxs[7] = mid_vtxs[1]

        mid_vtxs = BHL::Polyhedra.get_arc_mid_vtxs_prjct(@icosa_cim2v3_vtxs[3], @icosa_cim2v3_vtxs[9], 3, @radius)

        @icosa_cim2v3_vtxs[6] = mid_vtxs[0]
        @icosa_cim2v3_vtxs[8] = mid_vtxs[1]

      end # calc_icosa_cim2v3_vertices()

      def get_icosa_cim2v3_vertices()

        @icosa_cim2v3_vtxs

      end # get_icosa_cim2v3_vertices()

      def create_cim2v3_face_component(model)

        cmpnt_def = model.definitions.add

        cmpnt_def.name = "Icosa C I Method 2 V3 Face"

        @@face_vtx_map_v3.each do |face|

          cmpnt_def.entities.add_face(@icosa_cim2v3_vtxs[face[0]], @icosa_cim2v3_vtxs[face[1]], @icosa_cim2v3_vtxs[face[2]])

        end

        cmpnt_def

      end # create_cim2v3_face_component(model)

    end # class IcosahedronCIM2V3

    class IcosahedronCIM2V4 < Icosahedron

      def initialize(radius)

        super

        @icosa_vtxs = get_icosa_face_vertices

        calc_icosa_cim2v4_vertices()

      end # initialize(radius)

      def calc_icosa_cim2v4_vertices()

        @icosa_cim2v4_vtxs = Array.new(15)

        @icosa_cim2v4_vtxs[ 0] = @icosa_vtxs[0]
        @icosa_cim2v4_vtxs[ 4] = @icosa_vtxs[1]
        @icosa_cim2v4_vtxs[14] = @icosa_vtxs[2]

        # instead could get v2 vtxs and use mid points between those on ppt eges

        mid_vtxs = BHL::Polyhedra.get_arc_mid_vtxs_prjct(@icosa_cim2v4_vtxs[0], @icosa_cim2v4_vtxs[4], 4, @radius)

        @icosa_cim2v4_vtxs[1] = mid_vtxs[0]
        @icosa_cim2v4_vtxs[2] = mid_vtxs[1]
        @icosa_cim2v4_vtxs[3] = mid_vtxs[2]

        mid_vtxs = BHL::Polyhedra.get_arc_mid_vtxs_prjct(@icosa_cim2v4_vtxs[0], @icosa_cim2v4_vtxs[14], 4, @radius)

        @icosa_cim2v4_vtxs[ 5] = mid_vtxs[0]
        @icosa_cim2v4_vtxs[ 9] = mid_vtxs[1]
        @icosa_cim2v4_vtxs[12] = mid_vtxs[2]

        mid_vtxs = BHL::Polyhedra.get_arc_mid_vtxs_prjct(@icosa_cim2v4_vtxs[4], @icosa_cim2v4_vtxs[14], 4, @radius)

        @icosa_cim2v4_vtxs[ 8] = mid_vtxs[0]
        @icosa_cim2v4_vtxs[11] = mid_vtxs[1]
        @icosa_cim2v4_vtxs[13] = mid_vtxs[2]

        # method 2 mid face vertices on arcs between ppt edge vertices are not coincident
        # get vtxs from all 6 arcs and take the average at each vertex

        mid_face_vtxs = [[]]

        mid_face_vtxs[0] = BHL::Polyhedra.get_arc_mid_vtxs_prjct(@icosa_cim2v4_vtxs[5], @icosa_cim2v4_vtxs[ 8], 3, @radius)
        mid_face_vtxs[1] = BHL::Polyhedra.get_arc_mid_vtxs_prjct(@icosa_cim2v4_vtxs[9], @icosa_cim2v4_vtxs[11], 2, @radius)
        mid_face_vtxs[2] = BHL::Polyhedra.get_arc_mid_vtxs_prjct(@icosa_cim2v4_vtxs[1], @icosa_cim2v4_vtxs[13], 3, @radius)
        mid_face_vtxs[3] = BHL::Polyhedra.get_arc_mid_vtxs_prjct(@icosa_cim2v4_vtxs[2], @icosa_cim2v4_vtxs[11], 2, @radius)
        mid_face_vtxs[4] = BHL::Polyhedra.get_arc_mid_vtxs_prjct(@icosa_cim2v4_vtxs[3], @icosa_cim2v4_vtxs[12], 3, @radius)
        mid_face_vtxs[5] = BHL::Polyhedra.get_arc_mid_vtxs_prjct(@icosa_cim2v4_vtxs[2], @icosa_cim2v4_vtxs[ 9], 2, @radius)

#         0
#      2 / \ 0
#       1---5
#    3 / \ / \ 1
#   5 2---6---9
#    / \ / \ / \
# 4 3---7--10  12
#  / \ / \ / \ / \
# 4---8--11--13--14

        @icosa_cim2v4_vtxs[6] = 

          BHL::Polyhedra.get_3_vtx_avg_prjct(mid_face_vtxs[0][0], mid_face_vtxs[2][0], mid_face_vtxs[5][0], @radius)

        @icosa_cim2v4_vtxs[7] = 

          BHL::Polyhedra.get_3_vtx_avg_prjct(mid_face_vtxs[0][1], mid_face_vtxs[3][0], mid_face_vtxs[4][0], @radius)

        @icosa_cim2v4_vtxs[10] = 

          BHL::Polyhedra.get_3_vtx_avg_prjct(mid_face_vtxs[1][0], mid_face_vtxs[2][1], mid_face_vtxs[4][1], @radius)

      end # calc_icosa_cim2v4_vertices()

      def get_icosa_cim2v4_vertices()

        @icosa_cim2v4_vtxs

      end # get_icosa_cim2v4_vertices()

      def create_cim2v4_face_component(model)

        cmpnt_def = model.definitions.add

        cmpnt_def.name = "Icosa C I Method 2 V4 Face"

        @@face_vtx_map_v4.each do |face|

          cmpnt_def.entities.add_face(@icosa_cim2v4_vtxs[face[0]], @icosa_cim2v4_vtxs[face[1]], @icosa_cim2v4_vtxs[face[2]])

        end

        cmpnt_def

      end # create_cim2v4_face_component(model)

    end # class IcosahedronCIM2V4

    class IcosahedronCIM2V5 < Icosahedron

      def initialize(radius)

        super

        @icosa_vtxs = get_icosa_face_vertices

        calc_icosa_cim2v5_vertices()

      end # initialize(radius)

      def calc_icosa_cim2v5_vertices()

        @icosa_cim2v5_vtxs = Array.new(21)

        @icosa_cim2v5_vtxs[ 0] = @icosa_vtxs[0]
        @icosa_cim2v5_vtxs[ 5] = @icosa_vtxs[1]
        @icosa_cim2v5_vtxs[20] = @icosa_vtxs[2]

        mid_vtxs = BHL::Polyhedra.get_arc_mid_vtxs_prjct(@icosa_cim2v5_vtxs[0], @icosa_cim2v5_vtxs[5], 5, @radius)

        @icosa_cim2v5_vtxs[1] = mid_vtxs[0]
        @icosa_cim2v5_vtxs[2] = mid_vtxs[1]
        @icosa_cim2v5_vtxs[3] = mid_vtxs[2]
        @icosa_cim2v5_vtxs[4] = mid_vtxs[3]

        mid_vtxs = BHL::Polyhedra.get_arc_mid_vtxs_prjct(@icosa_cim2v5_vtxs[0], @icosa_cim2v5_vtxs[20], 5, @radius)

        @icosa_cim2v5_vtxs[ 6] = mid_vtxs[0]
        @icosa_cim2v5_vtxs[11] = mid_vtxs[1]
        @icosa_cim2v5_vtxs[15] = mid_vtxs[2]
        @icosa_cim2v5_vtxs[18] = mid_vtxs[3]

        mid_vtxs = BHL::Polyhedra.get_arc_mid_vtxs_prjct(@icosa_cim2v5_vtxs[5], @icosa_cim2v5_vtxs[20], 5, @radius)

        @icosa_cim2v5_vtxs[10] = mid_vtxs[0]
        @icosa_cim2v5_vtxs[14] = mid_vtxs[1]
        @icosa_cim2v5_vtxs[17] = mid_vtxs[2]
        @icosa_cim2v5_vtxs[19] = mid_vtxs[3]

        # method 2 mid face vertices on arcs between ppt edge vertices are not coincident
        # get vtxs from all 9 arcs and take the average at each vertex

        mid_face_vtxs = [[]]

        mid_face_vtxs[0] = BHL::Polyhedra.get_arc_mid_vtxs_prjct(@icosa_cim2v5_vtxs[ 6], @icosa_cim2v5_vtxs[10], 4, @radius)
        mid_face_vtxs[1] = BHL::Polyhedra.get_arc_mid_vtxs_prjct(@icosa_cim2v5_vtxs[11], @icosa_cim2v5_vtxs[14], 3, @radius)
        mid_face_vtxs[2] = BHL::Polyhedra.get_arc_mid_vtxs_prjct(@icosa_cim2v5_vtxs[15], @icosa_cim2v5_vtxs[17], 2, @radius)
        mid_face_vtxs[3] = BHL::Polyhedra.get_arc_mid_vtxs_prjct(@icosa_cim2v5_vtxs[ 1], @icosa_cim2v5_vtxs[19], 4, @radius)
        mid_face_vtxs[4] = BHL::Polyhedra.get_arc_mid_vtxs_prjct(@icosa_cim2v5_vtxs[ 2], @icosa_cim2v5_vtxs[17], 3, @radius)
        mid_face_vtxs[5] = BHL::Polyhedra.get_arc_mid_vtxs_prjct(@icosa_cim2v5_vtxs[ 3], @icosa_cim2v5_vtxs[14], 2, @radius)
        mid_face_vtxs[6] = BHL::Polyhedra.get_arc_mid_vtxs_prjct(@icosa_cim2v5_vtxs[ 4], @icosa_cim2v5_vtxs[18], 4, @radius)
        mid_face_vtxs[7] = BHL::Polyhedra.get_arc_mid_vtxs_prjct(@icosa_cim2v5_vtxs[ 3], @icosa_cim2v5_vtxs[15], 3, @radius)
        mid_face_vtxs[8] = BHL::Polyhedra.get_arc_mid_vtxs_prjct(@icosa_cim2v5_vtxs[ 2], @icosa_cim2v5_vtxs[11], 2, @radius)

#             0
#          3 / \ 0
#           1---6
#       4  / \ / \ 1
#       8 2---7---11
#      5 / \ / \ / \ 2
#     7 3---8--12--15
#      / \ / \ / \ / \
#   6 4---9--13--16--18
#    / \ / \ / \ / \ / \
#   5--10--14--17--19--20

        @icosa_cim2v5_vtxs[7] = 

          BHL::Polyhedra.get_3_vtx_avg_prjct(mid_face_vtxs[0][0], mid_face_vtxs[3][0], mid_face_vtxs[8][0], @radius)

        @icosa_cim2v5_vtxs[8] = 

          BHL::Polyhedra.get_3_vtx_avg_prjct(mid_face_vtxs[0][1], mid_face_vtxs[4][0], mid_face_vtxs[7][0], @radius)

        @icosa_cim2v5_vtxs[9] = 

          BHL::Polyhedra.get_3_vtx_avg_prjct(mid_face_vtxs[0][2], mid_face_vtxs[5][0], mid_face_vtxs[6][0], @radius)

        @icosa_cim2v5_vtxs[12] = 

          BHL::Polyhedra.get_3_vtx_avg_prjct(mid_face_vtxs[1][0], mid_face_vtxs[3][1], mid_face_vtxs[7][1], @radius)

        @icosa_cim2v5_vtxs[13] = 

          BHL::Polyhedra.get_3_vtx_avg_prjct(mid_face_vtxs[1][1], mid_face_vtxs[4][1], mid_face_vtxs[6][1], @radius)

        @icosa_cim2v5_vtxs[16] = 

          BHL::Polyhedra.get_3_vtx_avg_prjct(mid_face_vtxs[2][0], mid_face_vtxs[3][2], mid_face_vtxs[6][2], @radius)

      end # calc_icosa_cim2v5_vertices()

      def get_icosa_cim2v5_vertices()

        @icosa_cim2v5_vtxs

      end # get_icosa_cim2v5_vertices()

      def create_cim2v5_face_component(model)

        cmpnt_def = model.definitions.add

        cmpnt_def.name = "Icosa C I Method 2 V5 Face"

        @@face_vtx_map_v5.each do |face|

          cmpnt_def.entities.add_face(@icosa_cim2v5_vtxs[face[0]], @icosa_cim2v5_vtxs[face[1]], @icosa_cim2v5_vtxs[face[2]])

        end

        cmpnt_def

      end # create_cim2v5_face_component(model)

    end # class IcosahedronCIM2V5

    class IcosahedronCIM2V6 < Icosahedron

      def initialize(radius)

        super

        @icosa_vtxs = get_icosa_face_vertices

        calc_icosa_cim2v6_vertices()

      end # initialize(radius)

      def calc_icosa_cim2v6_vertices()

        @icosa_cim2v6_vtxs = Array.new(28)

        @icosa_cim2v6_vtxs[ 0] = @icosa_vtxs[0]
        @icosa_cim2v6_vtxs[ 6] = @icosa_vtxs[1]
        @icosa_cim2v6_vtxs[27] = @icosa_vtxs[2]

        @icosa_cim2v6_vtxs[15] = @icosa_vtxs[3]

        # could get m2 v3 vtxs and use mid points between those on ppt eges

        mid_vtxs = BHL::Polyhedra.get_arc_mid_vtxs_prjct(@icosa_cim2v6_vtxs[0], @icosa_cim2v6_vtxs[6], 6, @radius)

        @icosa_cim2v6_vtxs[1] = mid_vtxs[0]
        @icosa_cim2v6_vtxs[2] = mid_vtxs[1]
        @icosa_cim2v6_vtxs[3] = mid_vtxs[2]
        @icosa_cim2v6_vtxs[4] = mid_vtxs[3]
        @icosa_cim2v6_vtxs[5] = mid_vtxs[4]

        mid_vtxs = BHL::Polyhedra.get_arc_mid_vtxs_prjct(@icosa_cim2v6_vtxs[0], @icosa_cim2v6_vtxs[27], 6, @radius)

        @icosa_cim2v6_vtxs[ 7] = mid_vtxs[0]
        @icosa_cim2v6_vtxs[13] = mid_vtxs[1]
        @icosa_cim2v6_vtxs[18] = mid_vtxs[2]
        @icosa_cim2v6_vtxs[22] = mid_vtxs[3]
        @icosa_cim2v6_vtxs[25] = mid_vtxs[4]

        mid_vtxs = BHL::Polyhedra.get_arc_mid_vtxs_prjct(@icosa_cim2v6_vtxs[6], @icosa_cim2v6_vtxs[27], 6, @radius)

        @icosa_cim2v6_vtxs[12] = mid_vtxs[0]
        @icosa_cim2v6_vtxs[17] = mid_vtxs[1]
        @icosa_cim2v6_vtxs[21] = mid_vtxs[2]
        @icosa_cim2v6_vtxs[24] = mid_vtxs[3]
        @icosa_cim2v6_vtxs[26] = mid_vtxs[4]

        # method 2 mid face vertices on chords between ppt edge vertices are not coincident
        # get vtxs from all 12 chords and take the average at each vertex

        mid_face_vtxs = [[]]

        mid_face_vtxs[ 0] = BHL::Polyhedra.get_arc_mid_vtxs_prjct(@icosa_cim2v6_vtxs[ 7], @icosa_cim2v6_vtxs[12], 5, @radius)
        mid_face_vtxs[ 1] = BHL::Polyhedra.get_arc_mid_vtxs_prjct(@icosa_cim2v6_vtxs[13], @icosa_cim2v6_vtxs[17], 4, @radius)
        mid_face_vtxs[ 2] = BHL::Polyhedra.get_arc_mid_vtxs_prjct(@icosa_cim2v6_vtxs[18], @icosa_cim2v6_vtxs[21], 3, @radius)
        mid_face_vtxs[ 3] = BHL::Polyhedra.get_arc_mid_vtxs_prjct(@icosa_cim2v6_vtxs[22], @icosa_cim2v6_vtxs[24], 2, @radius)
        mid_face_vtxs[ 4] = BHL::Polyhedra.get_arc_mid_vtxs_prjct(@icosa_cim2v6_vtxs[ 1], @icosa_cim2v6_vtxs[26], 5, @radius)
        mid_face_vtxs[ 5] = BHL::Polyhedra.get_arc_mid_vtxs_prjct(@icosa_cim2v6_vtxs[ 2], @icosa_cim2v6_vtxs[24], 4, @radius)
        mid_face_vtxs[ 6] = BHL::Polyhedra.get_arc_mid_vtxs_prjct(@icosa_cim2v6_vtxs[ 3], @icosa_cim2v6_vtxs[21], 3, @radius)
        mid_face_vtxs[ 7] = BHL::Polyhedra.get_arc_mid_vtxs_prjct(@icosa_cim2v6_vtxs[ 4], @icosa_cim2v6_vtxs[17], 2, @radius)
        mid_face_vtxs[ 8] = BHL::Polyhedra.get_arc_mid_vtxs_prjct(@icosa_cim2v6_vtxs[ 5], @icosa_cim2v6_vtxs[25], 5, @radius)
        mid_face_vtxs[ 9] = BHL::Polyhedra.get_arc_mid_vtxs_prjct(@icosa_cim2v6_vtxs[ 4], @icosa_cim2v6_vtxs[22], 4, @radius)
        mid_face_vtxs[10] = BHL::Polyhedra.get_arc_mid_vtxs_prjct(@icosa_cim2v6_vtxs[ 3], @icosa_cim2v6_vtxs[18], 3, @radius)
        mid_face_vtxs[11] = BHL::Polyhedra.get_arc_mid_vtxs_prjct(@icosa_cim2v6_vtxs[ 2], @icosa_cim2v6_vtxs[13], 2, @radius)

#             0
#          4 / \ 0
#           1---7
#        5 / \ / \ 1
#      11 2---8---13
#      6 / \ / \ / \ 2
#    10 3---9--14--18
#    7 / \ / \ / \ / \ 3
#   9 4--10--15--19--22
#    / \ / \ / \ / \ / \
# 8 5--11--16--20--23--25
#  / \ / \ / \ / \ / \ / \
# 6--12--17--21--24--26--27

        @icosa_cim2v6_vtxs[8] = 

          BHL::Polyhedra.get_3_vtx_avg_prjct(mid_face_vtxs[0][0], mid_face_vtxs[4][0], mid_face_vtxs[11][0], @radius)

        @icosa_cim2v6_vtxs[9] = 

          BHL::Polyhedra.get_3_vtx_avg_prjct(mid_face_vtxs[0][1], mid_face_vtxs[5][0], mid_face_vtxs[10][0], @radius)

        @icosa_cim2v6_vtxs[10] = 

          BHL::Polyhedra.get_3_vtx_avg_prjct(mid_face_vtxs[0][2], mid_face_vtxs[6][0], mid_face_vtxs[9][0], @radius)

        @icosa_cim2v6_vtxs[11] = 

          BHL::Polyhedra.get_3_vtx_avg_prjct(mid_face_vtxs[0][3], mid_face_vtxs[7][0], mid_face_vtxs[8][0], @radius)

        @icosa_cim2v6_vtxs[14] = 

          BHL::Polyhedra.get_3_vtx_avg_prjct(mid_face_vtxs[1][0], mid_face_vtxs[4][1], mid_face_vtxs[10][1], @radius)

        @icosa_cim2v6_vtxs[16] = 

          BHL::Polyhedra.get_3_vtx_avg_prjct(mid_face_vtxs[1][2], mid_face_vtxs[6][1], mid_face_vtxs[8][1], @radius)

        @icosa_cim2v6_vtxs[19] = 

          BHL::Polyhedra.get_3_vtx_avg_prjct(mid_face_vtxs[2][0], mid_face_vtxs[4][2], mid_face_vtxs[9][2], @radius)

        @icosa_cim2v6_vtxs[20] = 

          BHL::Polyhedra.get_3_vtx_avg_prjct(mid_face_vtxs[2][1], mid_face_vtxs[5][2], mid_face_vtxs[8][2], @radius)

        @icosa_cim2v6_vtxs[23] = 

          BHL::Polyhedra.get_3_vtx_avg_prjct(mid_face_vtxs[3][0], mid_face_vtxs[4][3], mid_face_vtxs[8][3], @radius)

      end # calc_icosa_cim2v6_vertices()

      def get_icosa_cim2v6_vertices()

        @icosa_cim2v6_vtxs

      end # get_icosa_cim2v6_vertices()

      def create_cim2v6_face_component(model)

        cmpnt_def = model.definitions.add

        cmpnt_def.name = "Icosa C I Method 2 V6 Face"

        @@face_vtx_map_v6.each do |face|

          cmpnt_def.entities.add_face(@icosa_cim2v6_vtxs[face[0]], @icosa_cim2v6_vtxs[face[1]], @icosa_cim2v6_vtxs[face[2]])

        end

        cmpnt_def

      end # create_cim2v6_face_component(model)

    end # class IcosahedronCIM2V6

    class IcosahedronCIKV3 < Icosahedron

      def initialize(radius)

        super

        @icosa_vtxs = get_icosa_face_vertices

        calc_icosa_cikv3_vertices()

      end # initialize(radius)

      def calc_icosa_cikv3_vertices()

        # Find the surface vertex that is on the intersection of a lesser circle 
        # plane and a plane from the origin and two adjacent icosa vertices.
        # For V3, the lesser circle planes pass through the face centre.
        # V3 has two groups of symmetric planes per face, but each group is just
        # a refelection above / below an equatorial plane.

        # Get the vector of the intersection of the two planes. Find the point on the 
        # vector closest to the origin. That point on the vector to the origin is 
        # one side of a right triangle with hypoteneuse the radius. The length of the 
        # opposite side offset on the vector gives the vtx on the surface.
        # Reflection and rotation give the other vertices...

        @icosa_cikv3_vtxs = Array.new(10)

        @icosa_cikv3_vtxs[0] = @icosa_vtxs[0]
        @icosa_cikv3_vtxs[3] = @icosa_vtxs[1]
        @icosa_cikv3_vtxs[9] = @icosa_vtxs[2]

        @icosa_cikv3_vtxs[5] = @icosa_vtxs[3]

        trnsfrm_vtx = Geom::Transformation.new(ORIGIN, @icosa_vtxs[4], 2 * PI / 5)

        @lc_plane_vtx = []

        @lc_plane_vtx[0] = @icosa_cikv3_vtxs[5].transform(trnsfrm_vtx)
        @lc_plane_vtx[1] = @icosa_cikv3_vtxs[5].transform(trnsfrm_vtx.inverse)

        lc_plane =

          Geom.fit_plane_to_points([@icosa_cikv3_vtxs[5], @lc_plane_vtx[0], @lc_plane_vtx[1]])

        vtx_orgn_plane =

          Geom.fit_plane_to_points(ORIGIN, @icosa_cikv3_vtxs[0], @icosa_cikv3_vtxs[3])

        vctr_intrsct = Geom.intersect_plane_plane(lc_plane, vtx_orgn_plane)

        pnt_intrsct = ORIGIN.project_to_line(vctr_intrsct)

        adj_side = pnt_intrsct.distance(ORIGIN)

        offset = sin(acos(adj_side / @radius)) * @radius

        new_vctr = Geom::Vector3d.new(vctr_intrsct[1].x * - offset, 
                                      vctr_intrsct[1].y * - offset, 
                                      vctr_intrsct[1].z * - offset)

        @icosa_cikv3_vtxs[2] = pnt_intrsct.offset(new_vctr)

        @icosa_cikv3_vtxs[7] = [ - @icosa_cikv3_vtxs[2][0],
                                   @icosa_cikv3_vtxs[2][1],
                                   @icosa_cikv3_vtxs[2][2]]

        trnsfrm_vtxs = Geom::Transformation.new(ORIGIN, @icosa_cikv3_vtxs[5], 2 * PI / 3)

        @icosa_cikv3_vtxs[8] = @icosa_cikv3_vtxs[2].transform(trnsfrm_vtxs)
        @icosa_cikv3_vtxs[1] = @icosa_cikv3_vtxs[7].transform(trnsfrm_vtxs)

        @icosa_cikv3_vtxs[4] = @icosa_cikv3_vtxs[2].transform(trnsfrm_vtxs.inverse)
        @icosa_cikv3_vtxs[6] = @icosa_cikv3_vtxs[7].transform(trnsfrm_vtxs.inverse)

      end # calc_icosa_cikv3_vertices()

      def get_icosa_cikv3_vertices()

        @icosa_cikv3_vtxs

      end # get_icosa_cikv3_vertices()

      def create_cikv3_face_component(model)

        cmpnt_def = model.definitions.add

        cmpnt_def.name = "Icosa C I Kruschke V3 Face"

        @@face_vtx_map_v3.each do |face|

          cmpnt_def.entities.add_face(@icosa_cikv3_vtxs[face[0]], @icosa_cikv3_vtxs[face[1]], @icosa_cikv3_vtxs[face[2]])

        end

        cmpnt_def

      end # create_cikv3_face_component(model)

    end # class IcosahedronCIKV3

    class IcosahedronCIKV4 < Icosahedron

      def initialize(radius)

        super

        @icosa_vtxs = get_icosa_face_vertices

        @icosa_cim1v4 = IcosahedronCIM1V4.new(radius)

        @icosa_cim1v4_vtxs = @icosa_cim1v4.get_icosa_cim1v4_vertices

        calc_icosa_cikv4_vertices()

      end # initialize(radius)

      def calc_icosa_cikv4_vertices()

        # Method 1 V4 already has equatorial planes - use those vtxs for mid PPT 
        # edge and face centre triangle. Two face centre triangle vertices provide 
        # one edge of the next lesser circle plane.
        # V4 has three groups of symmetric planes per face, the equatorial plane 
        # and a refelection above / below the equatorial plane.

        # Method is same as V3. Get the vector of the inters@icosa_cim1v4_vtxsection of two planes. 
        # Find the point on the vector closest to the origin. Calculate the offset 
        # on the vector from that point for the vtx on the surface.
        # reflection and rotation give the other vertices...

        @icosa_cikv4_vtxs = Array.new(15)

        @icosa_cikv4_vtxs[ 0] = @icosa_cim1v4_vtxs[ 0]
        @icosa_cikv4_vtxs[ 2] = @icosa_cim1v4_vtxs[ 2]
        @icosa_cikv4_vtxs[ 4] = @icosa_cim1v4_vtxs[ 4]
        @icosa_cikv4_vtxs[ 6] = @icosa_cim1v4_vtxs[ 6]
        @icosa_cikv4_vtxs[ 7] = @icosa_cim1v4_vtxs[ 7]
        @icosa_cikv4_vtxs[ 9] = @icosa_cim1v4_vtxs[ 9]
        @icosa_cikv4_vtxs[10] = @icosa_cim1v4_vtxs[10]
        @icosa_cikv4_vtxs[11] = @icosa_cim1v4_vtxs[11]
        @icosa_cikv4_vtxs[14] = @icosa_cim1v4_vtxs[14]

        # transforms to get lesser circle plane points
        # @icosa_cikv4_vtxs[6] by - 144 about @icosa_cim1v4_vtxs[0]
        # @icosa_cikv4_vtxs[10] by 144 about @icosa_cim1v4_vtxs[14]

        @lc_plane_vtx = []

        trnsfrm_vtx = Geom::Transformation.new(ORIGIN, @icosa_cim1v4_vtxs[0], - 4 * PI / 5)

        @lc_plane_vtx[0] = @icosa_cikv4_vtxs[6].transform(trnsfrm_vtx)

        trnsfrm_vtx = Geom::Transformation.new(ORIGIN, @icosa_cim1v4_vtxs[14], 4 * PI / 5)

        @lc_plane_vtx[1] = @icosa_cikv4_vtxs[10].transform(trnsfrm_vtx)

        lc_plane =

          Geom.fit_plane_to_points(@icosa_cikv4_vtxs[6], @lc_plane_vtx[0], @lc_plane_vtx[1])

        vtx_orgn_plane =

          Geom.fit_plane_to_points(ORIGIN, @icosa_cikv4_vtxs[0], @icosa_cikv4_vtxs[4])

        vctr_intrsct = Geom.intersect_plane_plane(lc_plane, vtx_orgn_plane)

        pnt_intrsct = ORIGIN.project_to_line(vctr_intrsct)

        adj_side = pnt_intrsct.distance(ORIGIN)

        offset = sin(acos(adj_side / @radius)) * @radius

        # note vector is in opposite direction to v3. why?
        # order of points in plane... assume changes direction of plane normal

        new_vctr = Geom::Vector3d.new(vctr_intrsct[1].x * offset, 
                                      vctr_intrsct[1].y * offset, 
                                      vctr_intrsct[1].z * offset)

        @icosa_cikv4_vtxs[1] = pnt_intrsct.offset(new_vctr)

        @icosa_cikv4_vtxs[5] = [ - @icosa_cikv4_vtxs[1][0],
                                   @icosa_cikv4_vtxs[1][1],
                                   @icosa_cikv4_vtxs[1][2]]

        trnsfrm_vtxs = Geom::Transformation.new(ORIGIN, @icosa_vtxs[3], 2 * PI / 3)

        @icosa_cikv4_vtxs[ 8] = @icosa_cikv4_vtxs[1].transform(trnsfrm_vtxs)
        @icosa_cikv4_vtxs[ 3] = @icosa_cikv4_vtxs[5].transform(trnsfrm_vtxs)

        @icosa_cikv4_vtxs[12] = @icosa_cikv4_vtxs[1].transform(trnsfrm_vtxs.inverse)
        @icosa_cikv4_vtxs[13] = @icosa_cikv4_vtxs[5].transform(trnsfrm_vtxs.inverse)

      end # calc_icosa_cikv4_vertices()

      def get_icosa_cikv4_vertices()

        @icosa_cikv4_vtxs

      end # get_icosa_cikv4_vertices()

      def create_cikv4_face_component(model)

        cmpnt_def = model.definitions.add

        cmpnt_def.name = "Icosa C I Kruschke V4 Face"

        @@face_vtx_map_v4.each do |face|

          cmpnt_def.entities.add_face(@icosa_cikv4_vtxs[face[0]], @icosa_cikv4_vtxs[face[1]], @icosa_cikv4_vtxs[face[2]])

        end

#        cmpnt_def.entities.add_face(@icosa_cikv4_vtxs[6], @lc_plane_vtx[0], @lc_plane_vtx[1])

 #       cmpnt_def.entities.add_face(ORIGIN, @icosa_cikv4_vtxs[0], @icosa_cikv4_vtxs[4])

  #      cmpnt_def.entities.add_cpoint(@icosa_cikv4_vtxs[1])
   #     cmpnt_def.entities.add_cpoint(@icosa_cikv4_vtxs[5])

        cmpnt_def

      end # create_cikv4_face_component(model)

    end # class IcosahedronCIKV4


    def nrm_to_radius(vertex, radius)

      nrm_fctr = radius / vertex.distance(ORIGIN)

      vertex = vertex.to_a.collect! { |n| n * nrm_fctr }

    end # nrm_to_radius(vertex, radius)

    def get_3_vtx_avg_prjct(vtx_1, vtx_2, vtx_3, radius)

      vtx_mid = [(vtx_1[0] + vtx_2[0] + vtx_3[0]) / 3,
                 (vtx_1[1] + vtx_2[1] + vtx_3[1]) / 3,
                 (vtx_1[2] + vtx_2[2] + vtx_3[2]) / 3]

      vtx_mid = nrm_to_radius(vtx_mid, radius)

    end # get_3_vtx_avg_prjct(vtx_1, vtx_2, vtx_3, radius)

    def get_linear_cmbntn_prjct(vtx_1, vtx_2, weight, denom, radius)

      vtx_mid = Geom::Point3d.linear_combination((denom - weight.to_f) / denom, vtx_1, weight.to_f / denom, vtx_2)

      vtx_mid = nrm_to_radius(vtx_mid, radius)

    end # get_linear_cmbntn_prjct(vtx_1, vtx_2, n, radius)

    def get_arc_mid_vtxs_prjct(vtx_1, vtx_2, n, radius)

      # couldn't see a way in SU to return a point by angular 
      # division between 2 points and a common point

      # this is a nifty modified slerp found here

      # http://number-none.com/product/Understanding%20Slerp,%20Then%20Not%20Using%20It

      vctr_1 = Geom::Vector3d.new(vtx_1[0], vtx_1[1], vtx_1[2])
      vctr_2 = Geom::Vector3d.new(vtx_2[0], vtx_2[1], vtx_2[2])

      cntr_angle = vctr_1.angle_between(vctr_2)

      sub_angle = cntr_angle / n

      vctr_1.normalize!
      vctr_2.normalize!

      dot_pr = vctr_1.dot(vctr_2)

      arc_mid_vtxs = Array.new(n - 1)

      for idx in 1..n - 1

        vctr_3 = Geom::Vector3d.new(vctr_2[0] - vctr_1[0] * dot_pr, 
                                    vctr_2[1] - vctr_1[1] * dot_pr, 
                                    vctr_2[2] - vctr_1[2] * dot_pr)

        vctr_3.normalize!

        vtx_mid = []

        vtx_mid[0] = vctr_1.x * cos(sub_angle * idx) + vctr_3.x * sin(sub_angle * idx)
        vtx_mid[1] = vctr_1.y * cos(sub_angle * idx) + vctr_3.y * sin(sub_angle * idx)
        vtx_mid[2] = vctr_1.z * cos(sub_angle * idx) + vctr_3.z * sin(sub_angle * idx)

        vtx_mid = nrm_to_radius(vtx_mid, radius)

        arc_mid_vtxs[idx - 1] = vtx_mid

      end

      arc_mid_vtxs

    end # get_arc_mid_vtxs_prjct(vtx1, vtx2, weight, denom, radius)

  end # module Polyhedra

end # module BHL
